@@ -1,7 +1,6 @@
-from __future__ import print_function
-
+import dask as dsk  # type: ignore
 import numpy as np
-import dask as dsk
+
 from xgcm.duck_array_ops import concatenate
 
 
