@@ -4,16 +4,14 @@ import xarray as xr
 
 from xgcm.grid import Axis, Grid
 
-from .datasets import (  # noqa: F401
-    all_2d,
-    all_datasets,
-    datasets,
-    datasets_grid_metric,
-    nonperiodic_1d,
-    nonperiodic_2d,
-    periodic_1d,
-    periodic_2d,
-)
+from .datasets import all_2d  # noqa: F401
+from .datasets import all_datasets  # noqa: F401
+from .datasets import datasets  # noqa: F401
+from .datasets import datasets_grid_metric  # noqa: F401
+from .datasets import nonperiodic_1d  # noqa: F401
+from .datasets import nonperiodic_2d  # noqa: F401
+from .datasets import periodic_1d  # noqa: F401
+from .datasets import periodic_2d  # noqa: F401
 
 
 # helper function to produce axes from datasets
@@ -47,7 +45,9 @@ def test_extend_right_left(discontinuity, right):
 
 
 @pytest.mark.parametrize("fill_value", [0, 10, 20])
-@pytest.mark.parametrize("boundary", ["fill", "extend", "extrapolate"])
+@pytest.mark.parametrize(
+    "boundary", ["fill", "extend", pytest.param("extrapolate", marks=pytest.mark.xfail)]
+)
 @pytest.mark.parametrize("periodic", [True, False])
 @pytest.mark.parametrize("is_left_edge", [True, False])
 @pytest.mark.parametrize("boundary_discontinuity", [None, 360])
@@ -166,14 +166,14 @@ def test_axis_repr(all_datasets):
     # TODO: make this more complete
 
 
-def test_get_axis_coord(all_datasets):
+def test_get_position_name(all_datasets):
     ds, periodic, expected = all_datasets
     axis_objs = _get_axes(ds)
     for ax_name, axis in axis_objs.items():
         # create a dataarray with each axis coordinate
         for position, coord in axis.coords.items():
             da = 1 * ds[coord]
-            assert axis._get_axis_coord(da) == (position, coord)
+            assert axis._get_position_name(da) == (position, coord)
 
 
 def test_axis_wrap_and_replace_2d(periodic_2d):
@@ -581,7 +581,14 @@ def test_axis_errors():
 
 
 @pytest.mark.parametrize(
-    "boundary", [None, "fill", "extend", "extrapolate", {"X": "fill", "Y": "extend"}]
+    "boundary",
+    [
+        None,
+        "fill",
+        "extend",
+        pytest.param("extrapolate", marks=pytest.mark.xfail(strict=True)),
+        {"X": "fill", "Y": "extend"},
+    ],
 )
 @pytest.mark.parametrize("fill_value", [None, 0, 1.0])
 def test_grid_create(all_datasets, boundary, fill_value):
@@ -625,9 +632,9 @@ def test_create_grid_no_comodo(all_datasets):
 
 
 def test_grid_no_coords(periodic_1d):
-
+    """Ensure that you can use xgcm with Xarray datasets that don't have dimension coordinates."""
     ds, periodic, expected = periodic_1d
-    ds_nocoords = ds.drop_dims(list(ds.dims.keys()))
+    ds_nocoords = ds.drop_vars(list(ds.dims.keys()))
 
     coords = expected["axes"]
     grid = Grid(ds_nocoords, periodic=periodic, coords=coords)
@@ -665,9 +672,11 @@ def test_grid_ops(all_datasets):
                 da_interp = grid.interp(ds[varname], axis_name, boundary=boundary)
                 da_interp_ax = axis.interp(ds[varname], boundary=boundary)
                 assert da_interp.equals(da_interp_ax)
+
                 da_diff = grid.diff(ds[varname], axis_name, boundary=boundary)
                 da_diff_ax = axis.diff(ds[varname], boundary=boundary)
                 assert da_diff.equals(da_diff_ax)
+
                 if boundary is not None:
                     da_cumsum = grid.cumsum(ds[varname], axis_name, boundary=boundary)
                     da_cumsum_ax = axis.cumsum(ds[varname], boundary=boundary)
@@ -680,7 +689,7 @@ def test_grid_ops(all_datasets):
     "boundary",
     [
         "fill",
-        # "extrapolate", # do we not support extrapolation anymore?
+        pytest.param("extrapolate", marks=pytest.mark.xfail(strict=True)),
         "extend",
         {"X": "fill", "Y": "extend"},
         {"X": "extend", "Y": "fill"},
@@ -701,6 +710,34 @@ def test_multi_axis_input(all_datasets, func, periodic, boundary):
         xr.testing.assert_allclose(serial, full)
 
 
+@pytest.mark.parametrize(
+    "func",
+    ["interp", "max", "min", "diff", "cumsum"],
+)
+@pytest.mark.parametrize(
+    "boundary",
+    [
+        "fill",
+        pytest.param("extrapolate", marks=pytest.mark.xfail),
+        "extend",
+        {"X": "fill", "Y": "extend"},
+        {"X": "extend", "Y": "fill"},
+    ],
+)
+def test_dask_vs_eager(all_datasets, func, boundary):
+    ds, coords, metrics = datasets_grid_metric("C")
+    grid = Grid(ds, coords=coords)
+    grid_method = getattr(grid, func)
+    eager_result = grid_method(ds.tracer, "X", boundary=boundary)
+
+    ds = ds.chunk({"xt": 1, "yt": 1, "time": 1, "zt": 1})
+    grid = Grid(ds, coords=coords)
+    grid_method = getattr(grid, func)
+    dask_result = grid_method(ds.tracer, "X", boundary=boundary).compute()
+
+    xr.testing.assert_allclose(dask_result, eager_result)
+
+
 def test_grid_dict_input_boundary_fill(nonperiodic_1d):
     """Test axis kwarg input functionality using dict input"""
     ds, _, _ = nonperiodic_1d
@@ -746,6 +783,8 @@ def test_invalid_fill_value_error():
         "cumsum",
         "cumint",
         "derivative",
+        # TODO: we can get rid of many of these after the release. With the grid_ufunc logic many of these go through the same codepath
+        # e.g. diff/interp/min/max all are the same, so we can probably reduce this to diff, cumsum, integrate, derivative, cumint
     ],
 )
 @pytest.mark.parametrize("gridtype", ["B", "C"])
@@ -758,21 +797,34 @@ def test_keep_coords(funcname, gridtype):
         result = func(ds.tracer, axis_name)
         base_coords = list(result.dims)
         augmented_coords = [
-            c for c in ds.tracer.coords if set(ds[c].dims).issubset(result.dims)
+            c
+            for c in ds.coords
+            if set(ds[c].dims).issubset(result.dims) and c not in result.dims
         ]
+
         if funcname in ["integrate", "average"]:
             assert set(result.coords) == set(base_coords + augmented_coords)
         else:
             assert set(result.coords) == set(base_coords)
-        #
+
+        # TODO: why is the behavior different for integrate and average?
         if funcname not in ["integrate", "average"]:
             result = func(ds.tracer, axis_name, keep_coords=False)
             assert set(result.coords) == set(base_coords)
-            #
+
             result = func(ds.tracer, axis_name, keep_coords=True)
             assert set(result.coords) == set(base_coords + augmented_coords)
 
 
+def test_keep_coords_deprecation():
+    ds, coords, metrics = datasets_grid_metric("B")
+    ds = ds.assign_coords(yt_bis=ds["yt"], xt_bis=ds["xt"])
+    grid = Grid(ds, coords=coords, metrics=metrics)
+    for axis_name in grid.axes.keys():
+        with pytest.warns(DeprecationWarning):
+            grid.diff(ds.tracer, axis_name, keep_coords=False)
+
+
 def test_boundary_kwarg_same_as_grid_constructor_kwarg():
     ds = datasets["2d_left"]
     grid1 = Grid(ds, periodic=False)
@@ -782,3 +834,283 @@ def test_boundary_kwarg_same_as_grid_constructor_kwarg():
     actual2 = grid2.interp(ds.data_g, ("X", "Y"))
 
     xr.testing.assert_identical(actual1, actual2)
+
+
+@pytest.mark.parametrize(
+    "metric_axes,metric_name",
+    [
+        (["Y", "X"], "area_n"),
+        ("X", "dx_t"),
+        ("Y", "dy_ne"),
+        (["Y", "X"], "dy_n"),
+        (["X"], "tracer"),
+    ],
+)
+@pytest.mark.parametrize("periodic", [True, False])
+@pytest.mark.parametrize(
+    "boundary, boundary_expected",
+    [
+        ({"X": "fill", "Y": "fill"}, {"X": "fill", "Y": "fill"}),
+        ({"X": "extend", "Y": "extend"}, {"X": "extend", "Y": "extend"}),
+        pytest.param(
+            {"X": "extrapolate", "Y": "extrapolate"},
+            {"X": "extrapolate", "Y": "extrapolate"},
+            marks=pytest.mark.xfail(
+                reason="padding via extrapolation not yet supported in grid_ufunc refactor"
+            ),
+        ),
+        ("fill", {"X": "fill", "Y": "fill"}),
+        ("extend", {"X": "extend", "Y": "extend"}),
+        pytest.param(
+            "extrapolate",
+            {"X": "extrapolate", "Y": "extrapolate"},
+            marks=pytest.mark.xfail(
+                reason="padding via extrapolation not yet supported in grid_ufunc refactor"
+            ),
+        ),
+        ({"X": "extend", "Y": "fill"}, {"X": "extend", "Y": "fill"}),
+        pytest.param(
+            {"X": "extrapolate", "Y": "fill"},
+            {"X": "extrapolate", "Y": "fill"},
+            marks=pytest.mark.xfail(
+                reason="padding via extrapolation not yet supported in grid_ufunc refactor"
+            ),
+        ),
+        pytest.param(
+            "fill",
+            {"X": "fill", "Y": "extend"},
+            marks=pytest.mark.xfail,
+            id="boundary not equal to boundary_expected",
+        ),
+    ],
+)
+@pytest.mark.parametrize("fill_value", [None, 0.1])
+def test_interp_like(
+    metric_axes, metric_name, periodic, boundary, boundary_expected, fill_value
+):
+
+    ds, coords, _ = datasets_grid_metric("C")
+    grid = Grid(ds, coords=coords, periodic=periodic)
+    grid.set_metrics(metric_axes, metric_name)
+    metric_available = grid._metrics.get(frozenset(metric_axes), None)
+    metric_available = metric_available[0]
+    interp_metric = grid.interp_like(
+        metric_available, ds.u, boundary=boundary, fill_value=fill_value
+    )
+    expected_metric = grid.interp(
+        ds[metric_name], metric_axes, boundary=boundary_expected, fill_value=fill_value
+    )
+
+    xr.testing.assert_allclose(interp_metric, expected_metric)
+
+
+def test_input_not_dims():
+    data = np.random.rand(4, 5)
+    coord = np.random.rand(4, 5)
+    ds = xr.DataArray(
+        data, dims=["x", "y"], coords={"c": (["x", "y"], coord)}
+    ).to_dataset(name="data")
+    msg = r"is not a dimension in the input dataset"
+    with pytest.raises(ValueError, match=msg):
+        Grid(ds, coords={"X": {"center": "c"}})
+
+
+def test_input_dim_notfound():
+    data = np.random.rand(4, 5)
+    coord = np.random.rand(4, 5)
+    ds = xr.DataArray(
+        data, dims=["x", "y"], coords={"c": (["x", "y"], coord)}
+    ).to_dataset(name="data")
+    msg = r"Could not find dimension `other` \(for the `center` position on axis `X`\) in input dataset."
+    with pytest.raises(ValueError, match=msg):
+        Grid(ds, coords={"X": {"center": "other"}})
+
+
+@pytest.mark.parametrize(
+    "funcname",
+    [
+        "interp",
+        "diff",
+        "min",
+        "max",
+        "cumsum",
+        "derivative",
+        "cumint",
+    ],
+)
+@pytest.mark.parametrize(
+    "boundary",
+    ["fill", "extend"],
+)
+@pytest.mark.parametrize(
+    "fill_value",
+    [0, 10, None],
+)
+def test_boundary_global_input(funcname, boundary, fill_value):
+    """Test that globally defined boundary values result in
+    the same output as when the parameters are defined the grid methods
+    """
+    ds, coords, metrics = datasets_grid_metric("C")
+    axis = "X"
+    # Test results by globally specifying fill value/boundary on grid object
+    grid_global = Grid(
+        ds,
+        coords=coords,
+        metrics=metrics,
+        periodic=False,
+        boundary=boundary,
+        fill_value=fill_value,
+    )
+    func_global = getattr(grid_global, funcname)
+    global_result = func_global(ds.tracer, axis)
+
+    # Test results by manually specifying fill value/boundary on grid method
+    grid_manual = Grid(
+        ds, coords=coords, metrics=metrics, periodic=False, boundary=boundary
+    )
+    func_manual = getattr(grid_manual, funcname)
+    manual_result = func_manual(
+        ds.tracer, axis, boundary=boundary, fill_value=fill_value
+    )
+    xr.testing.assert_allclose(global_result, manual_result)
+
+
+class TestInputErrorGridMethods:
+    def test_multiple_keys_vector_input(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Vector components provided as dictionaries should contain exactly one key/value pair. .*?"
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            grid.diff({"X": xr.DataArray(), "Y": xr.DataArray()}, "X")
+
+    def test_wrong_input_type_scalar(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "All data arguments must be either a DataArray or Dictionary .*?"
+        with pytest.raises(
+            TypeError,
+            match=msg,
+        ):
+            grid.diff("not_a_dataarray", "X")
+
+    def test_wrong_input_type_vector(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Dictionary inputs must have a DataArray as value. Got .*?"
+        with pytest.raises(
+            TypeError,
+            match=msg,
+        ):
+            grid.diff({"X": "not_a_dataarray"}, "X")
+
+    def test_wrong_axis_vector_input_axis(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Vector component with unknown axis provided. Grid has axes .*?"
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            grid.diff({"wrong": xr.DataArray()}, "X")
+
+
+class TestInputErrorApplyAsGridUfunc:
+    def test_multiple_keys_vector_input(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Vector components provided as dictionaries should contain exactly one key/value pair. .*?"
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            grid.apply_as_grid_ufunc(
+                lambda x: x, {"X": xr.DataArray(), "Y": xr.DataArray()}, "X"
+            )
+
+    def test_wrong_input_type_scalar(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "All data arguments must be either a DataArray or Dictionary .*?"
+        with pytest.raises(
+            TypeError,
+            match=msg,
+        ):
+            grid.apply_as_grid_ufunc(lambda x: x, "not_a_dataarray", "X")
+
+    def test_wrong_input_type_vector(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Dictionary inputs must have a DataArray as value. Got .*?"
+        with pytest.raises(
+            TypeError,
+            match=msg,
+        ):
+            grid.apply_as_grid_ufunc(lambda x: x, {"X": "not_a_dataarray"}, "X")
+
+    def test_wrong_axis_vector_input_axis(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Vector component with unknown axis provided. Grid has axes .*?"
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            grid.apply_as_grid_ufunc(lambda x: x, {"wrong": xr.DataArray()}, "X")
+
+    def test_vector_input_data_other_mismatch(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = (
+            "When providing multiple input arguments, `other_component`"
+            " needs to provide one dictionary per input"
+        )
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            # Passing 3 args and 2 other components should fail.
+            grid.apply_as_grid_ufunc(
+                lambda x: x,
+                {"X": xr.DataArray()},
+                {"Y": xr.DataArray()},
+                {"Z": xr.DataArray()},
+                axis="X",
+                other_component=[{"X": xr.DataArray()}, {"Y": xr.DataArray()}],
+            )
+
+    def test_wrong_input_type_vector_multi_input(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Dictionary inputs must have a DataArray as value. Got .*?"
+        with pytest.raises(
+            TypeError,
+            match=msg,
+        ):
+            # Passing 3 args and 2 other components should fail.
+            grid.apply_as_grid_ufunc(
+                lambda x: x,
+                {"X": xr.DataArray()},
+                {"Y": "not_a_data_array"},
+                axis="X",
+                other_component=[{"X": xr.DataArray()}, {"Y": xr.DataArray()}],
+            )
+
+    def test_wrong_axis_vector_input_axis_multi_input(self):
+        ds, _, _ = datasets_grid_metric("C")
+        grid = Grid(ds)
+        msg = "Vector component with unknown axis provided. Grid has axes .*?"
+        with pytest.raises(
+            ValueError,
+            match=msg,
+        ):
+            # Passing 3 args and 2 other components should fail.
+            grid.apply_as_grid_ufunc(
+                lambda x: x,
+                {"X": xr.DataArray()},
+                {"Y": xr.DataArray()},
+                axis="X",
+                other_component=[{"wrong": xr.DataArray()}, {"Y": xr.DataArray()}],
+            )
